# Copyright (c) 2017-2020 Patricio Cubillos and contributors.
# repack is open-source software under the MIT license (see LICENSE).

# repack version:
repack_VER =  1  # Major version
repack_MIN =  4  # Minor version
repack_REV =  0  # Revision
